/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/darwin/time.c: Timekeeping functions for Darwin-based systems
 * (Mac/iOS).
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/sysdep.h"
#include "src/sysdep/darwin/time.h"

#include <mach/mach.h>
#include <mach/mach_time.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Divisor for system-reported time (just in case Darwin someday gets
 * sub-nanosecond resolution). */
static uint8_t time_shift = 0;

/* sys_time_now() time units per second, as returned from sys_time_unit(). */
static unsigned int ticks_per_sec;

/* Last value returned from mach_absolute_time(). */
static uint64_t last_time;

/* Wraparound offset in case we're shifting the raw time values. */
static uint64_t time_offset = 0;

/* time_now() epoch in the mach_absolute_now() time base (expressed in
 * seconds). */
static double epoch = 0;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void sys_time_init(void)
{
    mach_timebase_info_data_t timebase_info;
    mach_timebase_info(&timebase_info);
    uint64_t ticks_per_sec_l = ((1000000000 * (uint64_t)timebase_info.denom
                                 + (uint64_t)(timebase_info.numer / 2))
                                / (uint64_t)timebase_info.numer);
    time_shift = 0;
    while (ticks_per_sec_l >= (uint64_t)1 << 31) {
        time_shift++;
        ticks_per_sec_l >>= 1;
    }
    ticks_per_sec = (int)ticks_per_sec_l;
    last_time = mach_absolute_time();
    epoch = 0;
}

/*-----------------------------------------------------------------------*/

int sys_time_unit(void)
{
    return ticks_per_sec;
}

/*-----------------------------------------------------------------------*/

uint64_t sys_time_now(void)
{
    const uint64_t now = mach_absolute_time();
    if (UNLIKELY(now < last_time) && time_shift > 0) {
        time_offset += (uint64_t)1 << (64 - time_shift);
    }
    const uint64_t time = (now >> time_shift) + time_offset;
    if (UNLIKELY(epoch == 0)) {
        epoch = (double)time / (double)ticks_per_sec;
    }
    return time;
}

/*-----------------------------------------------------------------------*/

void sys_time_delay(int64_t time)
{
    const uint64_t start = mach_absolute_time();
    const uint64_t target = start + (time << time_shift);
    if (start > target) {
        mach_wait_until(0);
    }
    uint64_t now;
    while ((now = mach_absolute_time()) >= start && now < target) {
        mach_wait_until(target);
    }
}

/*************************************************************************/
/*********************** Library-internal routines ***********************/
/*************************************************************************/

double darwin_time_epoch(void)
{
    return epoch;
}

/*-----------------------------------------------------------------------*/

double darwin_time_from_timestamp(uint64_t timestamp)
{
    timestamp = (timestamp >> time_shift) + time_offset;
    return (double)timestamp / (double)ticks_per_sec - epoch;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
